#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pytest

from pysolorie import AtmosphericTransmission


@pytest.mark.parametrize(
    "climate_type,"
    + "observer_altitude,"
    + "observer_latitude,"
    + "day_of_year,"
    + "solar_time,"
    + "expected_transmittance",
    [
        ("MIDLATITUDE SUMMER", 1200, 35.69, 81, 12 * 60 * 60, 0.683),  # Tehran Summer
        ("MIDLATITUDE WINTER", 1200, 35.69, 355, 12 * 60 * 60, 0.618),  # Tehran Winter
        ("TROPICAL", 63, 3.59, 81, 10 * 60 * 60, 0.597),  # Medan
        ("SUBARCTIC SUMMER", 136, 64.84, 1, 13 * 60 * 60, 0.140),  # Fairbanks
    ],
)
def test_calculate_transmittance(
    climate_type: str,
    observer_altitude: int,
    observer_latitude: float,
    day_of_year: int,
    solar_time: float,
    expected_transmittance: float,
) -> None:
    atmospheric_transmission: AtmosphericTransmission = AtmosphericTransmission(
        climate_type, observer_altitude, observer_latitude
    )
    result: float = atmospheric_transmission.calculate_transmittance(
        day_of_year, solar_time
    )
    assert pytest.approx(result, abs=1e-3) == expected_transmittance



# Copyright 2023 Alireza Aghamohammadi

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import math

from .model import HottelModel
from .observer import Observer


class AtmosphericTransmission:
    r"""
    A class to model the atmospheric transmission.
    """

    def __init__(
        self,
        climate_type: str,
        observer_altitude: int,
        observer_latitude: float,
    ):
        r"""
        To instantiate the ``AtmosphericTransmission`` class,
        provide the following three parameters.


        :param climate_type: The type of climate.
        :type climate_type: str
        :param observer_altitude: The altitude of the observer in meters.
        :type observer_altitude: int
        :param observer_latitude: The latitude of the observer in degrees.
        :type observer_latitude: float
        """
        self.hottel_model = HottelModel()
        (
            self.a0,
            self.a1,
            self.k,
        ) = self.hottel_model.calculate_transmittance_components(
            climate_type, observer_altitude
        )
        self.observer = Observer(observer_latitude)

    def calculate_transmittance(self, day_of_year: int, solar_time: float) -> float:
        r"""
        Calculate the effective atmospheric transmission coefficient of the direct beam.

        The effective atmospheric transmission coefficient of the direct beam
        is calculated using the formula:

        .. math::
            \tau_b = a_0 + a_1 \times \exp\left(-\frac{k}{\cos(\theta_z)}\right)


        | - :math:`\tau_b` is the effective atmospheric transmission
            coefficient of the direct beam
        | - :math:`a_0`, :math:`a_1`, and :math:`k` are the components
            of clear-sky beam radiation transmittance
        | - :math:`\theta_z` is the solar zenith angle

        :param day_of_year: The day of the year.
        :type day_of_year: int
        :param solar_time: The solar time in seconds.
        :type solar_time: float
        :return: The effective atmospheric transmission coefficient of the direct beam.
        :rtype: float
        """

        zenith_angle = self.observer.calculate_zenith_angle(day_of_year, solar_time)
        cos_zenith_angle = math.cos(zenith_angle)
        EPSILON = 1e-8  # Small constant to prevent division by zero
        if abs(cos_zenith_angle) < EPSILON:
            return self.a0
        else:
            return self.a0 + self.a1 * math.exp(-self.k / cos_zenith_angle)



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import math


class SunPosition:
    r"""
    A class to model sun position.
    """

    def solar_declination(self, day_of_year: int) -> float:
        r"""
        Calculate the solar declination angle in radians [1]_.

        The solar declination angle is the angle between the rays of the sun
        and the plane of the Earth's equator.

        The formula used to calculate the solar declination angle is:

        .. math::
            \delta = \frac{23.45 \pi}{180} \times
            \sin \left(\frac{2\pi~(284 + n)}{365}\right)

        | :math:`n` is the day of the year (i.e., ``day_of_year``)

        :param day_of_year: The day of the year.
        :type day_of_year: int
        :return: The solar declination angle in radians.
        :rtype: float

        References
        ----------
        .. [1] Cooper, P. (1969). The absorption of radiation in solar stills.
                Solar Energy, 12(3), 333-346.
        """
        # tilt of the Earth's axis (in degrees)
        earth_tilt_degrees = 23.45

        # Convert the tilt to radians
        earth_tilt_radians = math.radians(earth_tilt_degrees)

        # Offset to ensure declination angle is zero at the March equinox
        equinox_offset_days = 284

        # Calculate the solar declination angle
        solar_declination = (
            math.sin((2 * math.pi) * (equinox_offset_days + day_of_year) / 365)
            * earth_tilt_radians
        )

        return solar_declination

    def hour_angle(self, solar_time: float) -> float:
        r"""
        Calculate the hour angle based on the solar time.

        The hour angle is a measure of time, expressed in angular terms,
        from solar noon.

        The formula used to calculate the hour angle is:

        .. math::
            \omega = (t - 12\times 60 \times 60)
            \times \frac{\pi}{12\times 60 \times 60}

        :param solar_time: The solar time in seconds.
        :type solar_time: float
        :return: The hour angle in radians.
        :rtype: float
        """
        # The number of seconds in half a day (12 hours)
        seconds_in_half_day = 12 * 60 * 60

        # The Earth rotates by pi/seconds_in_half_day radians per second
        earth_rotation_rate = math.pi / seconds_in_half_day

        # Calculate the hour angle
        hour_angle = (solar_time - seconds_in_half_day) * earth_rotation_rate

        return hour_angle

    def solar_time(self, hour_angle: float) -> float:
        r"""
        Calculate the solar time based on the hour angle.

        The solar time is a measure of time, expressed in seconds, from solar noon.

        The formula used to calculate the solar time is:

        .. math::
            t = \omega \times \frac{12 \times 60 \times 60}{\pi}
            + 12 \times 60 \times 60

        :param hour_angle: The hour angle in radians.
        :type hour_angle: float
        :return: The solar time in seconds.
        :rtype: float
        """
        # The number of seconds in half a day (12 hours)
        seconds_in_half_day = 12 * 60 * 60

        # The Earth rotates by pi/seconds_in_half_day radians per second
        earth_rotation_rate = math.pi / seconds_in_half_day

        # Calculate the solar time
        solar_time = hour_angle / earth_rotation_rate + seconds_in_half_day

        return solar_time



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import csv
import json
import logging
import xml.etree.ElementTree as ET
from pathlib import Path

import pytest

from pysolorie import IrradiationCalculator, ReportGenerator


def test_generate_optimal_orientation_csv_report(caplog, tmpdir) -> None:
    caplog.set_level(logging.INFO)

    # Create a temporary directory for the test
    temp_dir: Path = Path(tmpdir)

    # Initialize the ReportGenerator
    report_generator: ReportGenerator = ReportGenerator()

    # Initialize the IrradiationCalculator for Tehran
    irradiation_calculator: IrradiationCalculator = IrradiationCalculator(
        "MIDLATITUDE SUMMER", 1200, 35.6892
    )

    # Define the path for the CSV file
    csv_path: Path = temp_dir / "report.csv"
    from_day: int = 60
    to_day: int = 70
    # Call the method to generate the report
    report_generator.generate_optimal_orientation_csv_report(
        csv_path, irradiation_calculator, from_day, to_day
    )

    # Check the CSV file
    with open(csv_path, "r") as file:
        reader = csv.reader(file)
        header = next(reader)
        assert header == [
            "Day",
            "Beta (degrees)",
            "Direct Irradiation (Megajoules per square meter)",
        ]
        for i, row in enumerate(reader, start=from_day):
            day, beta, total_direct_irradiation = (
                int(row[0]),
                float(row[1]),
                float(row[2]),
            )

            assert day == i
            expected_beta = irradiation_calculator.find_optimal_orientation(i)

            expected_total_direct_irradiation = (
                irradiation_calculator.calculate_direct_irradiation(beta, i)
            )
            assert pytest.approx(beta, abs=1e-3) == expected_beta
            assert (
                pytest.approx(total_direct_irradiation, abs=1e-3)
                == expected_total_direct_irradiation
            )
    # Check the logs
    for day in range(from_day, to_day):
        assert any(
            f"On day {day}," in record.message for record in caplog.records
        ), f"No log message for day {day}"


def test_generate_optimal_orientation_json_report(caplog, tmpdir) -> None:
    caplog.set_level(logging.INFO)

    # Create a temporary directory for the test
    temp_dir: Path = Path(tmpdir)

    # Initialize the ReportGenerator
    report_generator: ReportGenerator = ReportGenerator()

    # Initialize the IrradiationCalculator for Tehran
    irradiation_calculator: IrradiationCalculator = IrradiationCalculator(
        "MIDLATITUDE SUMMER", 1200, 35.6892
    )

    # Define the path for the JSON file
    json_path: Path = temp_dir / "report.json"
    from_day: int = 60
    to_day: int = 70
    # Call the method to generate the report
    report_generator.generate_optimal_orientation_json_report(
        json_path, irradiation_calculator, from_day, to_day
    )

    # Check the JSON file
    with open(json_path, "r") as file:
        data = json.load(file)
        for i, row in enumerate(data, start=from_day):
            day, beta, direct_irradiation = (
                row["Day"],
                row["Beta (degrees)"],
                row["Direct Irradiation (Megajoules per square meter)"],
            )

            assert day == i
            expected_beta = irradiation_calculator.find_optimal_orientation(i)
            expected_direct_irradiation = (
                irradiation_calculator.calculate_direct_irradiation(beta, i)
            )
            assert pytest.approx(beta, abs=1e-3) == expected_beta
            assert (
                pytest.approx(direct_irradiation, abs=1e-3)
                == expected_direct_irradiation
            )
    # Check the logs
    for day in range(from_day, to_day):
        assert any(
            f"On day {day}," in record.message for record in caplog.records
        ), f"No log message for day {day}"


def test_generate_optimal_orientation_xml_report(caplog, tmpdir) -> None:
    caplog.set_level(logging.INFO)

    # Create a temporary directory for the test
    temp_dir: Path = Path(tmpdir)

    # Initialize the ReportGenerator
    report_generator: ReportGenerator = ReportGenerator()

    # Initialize the IrradiationCalculator for Tehran
    irradiation_calculator: IrradiationCalculator = IrradiationCalculator(
        "MIDLATITUDE SUMMER", 1200, 35.6892
    )

    # Define the path for the XML file
    xml_path: Path = temp_dir / "report.xml"
    from_day: int = 60
    to_day: int = 70
    # Call the method to generate the report
    report_generator.generate_optimal_orientation_xml_report(
        xml_path, irradiation_calculator, from_day, to_day
    )

    # Check the XML file
    tree = ET.parse(xml_path)
    root = tree.getroot()

    for i, day_element in enumerate(root.findall("Day"), start=from_day):
        day = int(day_element.get("id"))
        beta = float(day_element.find("Beta").text)
        total_direct_irradiation = float(day_element.find("DirectIrradiation").text)

        assert day == i
        expected_beta = irradiation_calculator.find_optimal_orientation(i)
        expected_total_direct_irradiation = (
            irradiation_calculator.calculate_direct_irradiation(beta, i)
        )
        assert pytest.approx(beta, abs=1e-3) == expected_beta
        assert (
            pytest.approx(total_direct_irradiation, abs=1e-3)
            == expected_total_direct_irradiation
        )
    # Check the logs
    for day in range(from_day, to_day):
        assert any(
            f"On day {day}," in record.message for record in caplog.records
        ), f"No log message for day {day}"



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import csv
import json
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Dict, List, Union

from .logger import logger_decorator
from .numerical_integration import IrradiationCalculator


class ReportGenerator:
    r"""
    A class to generate reports for the optimal orientation
    of solar panels and the total direct irradiation.
    """

    @logger_decorator
    def _calculate_optimal_orientation_and_irradiation(
        self,
        irradiation_calculator: IrradiationCalculator,
        from_day: int,
        to_day: int,
    ) -> List[Dict[str, Union[int, float]]]:
        r"""
        This private method calculates the optimal solar panel orientation and
        total direct irradiation for a range of days.

        :param irradiation_calculator: An instance of the IrradiationCalculator class.
        :type irradiation_calculator: pysolorie.IrradiationCalculator
        :param from_day: The starting day of the report.
        :type from_day: int
        :param to_day: The ending day of the report.
        :type to_day: int
        :return: A list of dictionaries, each containing the day,
                 optimal orientation (beta),
                 and total direct irradiation.
        :rtype: List[Dict[str, Union[int, float]]]
        """
        data = []

        for day in range(from_day, to_day):
            beta = irradiation_calculator.find_optimal_orientation(day)
            total_direct_irradiation = (
                irradiation_calculator.calculate_direct_irradiation(beta, day)
            )

            self.logger.info(  # type: ignore
                f"On day {day}, the solar panel's optimal orientation is "
                f"{beta} degrees, and the total direct irradiation is "
                f"{total_direct_irradiation} Megajoules per square meter."
            )

            # Append the result to the data list
            data.append(
                {
                    "Day": day,
                    "Beta (degrees)": beta,
                    "Direct Irradiation "
                    "(Megajoules per square meter)": total_direct_irradiation,
                }
            )

        return data

    @logger_decorator
    def generate_optimal_orientation_csv_report(
        self,
        path: Path,
        irradiation_calculator: IrradiationCalculator,
        from_day: int,
        to_day: int,
    ) -> None:
        r"""
        This method generates a report of optimal solar panel orientation in CSV format.
        It uses the ``_calculate_optimal_orientation_and_irradiation``
        method to get the data.

        :param path: A Path object that points to the CSV file
                    where the report will be written.
        :type path: Path
        :param irradiation_calculator: An instance of the IrradiationCalculator class.
        :type irradiation_calculator: pysolorie.IrradiationCalculator
        :param from_day: The starting day of the report.
        :type from_day: int
        :param to_day: The ending day of the report.
        :type to_day: int
        """
        data = self._calculate_optimal_orientation_and_irradiation(
            irradiation_calculator, from_day, to_day
        )

        with open(path, "w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow(
                [
                    "Day",
                    "Beta (degrees)",
                    "Direct Irradiation (Megajoules per square meter)",
                ]
            )

            for row in data:
                writer.writerow(
                    [
                        row["Day"],
                        row["Beta (degrees)"],
                        row["Direct Irradiation (Megajoules per square meter)"],
                    ]
                )

    @logger_decorator
    def generate_optimal_orientation_json_report(
        self,
        path: Path,
        irradiation_calculator: IrradiationCalculator,
        from_day: int,
        to_day: int,
    ) -> None:
        r"""
        This method generates a report of optimal
        solar panel orientation in JSON format.
        It uses the ``_calculate_optimal_orientation_and_irradiation``
        method to get the data.

        :param path: A Path object that points to the JSON file
                    where the report will be written.
        :type path: Path
        :param irradiation_calculator: An instance of the IrradiationCalculator class.
        :type irradiation_calculator: pysolorie.IrradiationCalculator
        :param from_day: The starting day of the report.
        :type from_day: int
        :param to_day: The ending day of the report.
        :type to_day: int
        """
        data = self._calculate_optimal_orientation_and_irradiation(
            irradiation_calculator, from_day, to_day
        )

        # Write the data list to the JSON file
        with open(path, "w") as file:
            json.dump(data, file, indent=4)

    @logger_decorator
    def generate_optimal_orientation_xml_report(
        self,
        path: Path,
        irradiation_calculator: IrradiationCalculator,
        from_day: int,
        to_day: int,
    ) -> None:
        r"""
        This method generates a report of optimal solar panel orientation in XML format.
        It uses the ``_calculate_optimal_orientation_and_irradiation``
        method to get the data.

        :param path: A Path object that points to the XML file
                    where the report will be written.
        :type path: Path
        :param irradiation_calculator: An instance of the IrradiationCalculator class.
        :type irradiation_calculator: pysolorie.IrradiationCalculator
        :param from_day: The starting day of the report.
        :type from_day: int
        :param to_day: The ending day of the report.
        :type to_day: int
        """
        data = self._calculate_optimal_orientation_and_irradiation(
            irradiation_calculator, from_day, to_day
        )

        # Create the root element
        root = ET.Element("Report")

        for row in data:
            # Create a 'Day' element for each day
            day_element = ET.SubElement(root, "Day")
            day_element.set("id", str(row["Day"]))

            # Create 'Beta' and 'DirectIrradiation' elements for each day
            beta_element = ET.SubElement(day_element, "Beta")
            beta_element.text = str(row["Beta (degrees)"])
            tdi_element = ET.SubElement(day_element, "DirectIrradiation")
            tdi_element.text = str(
                row["Direct Irradiation (Megajoules per square meter)"]
            )

        # Write the XML data to the file
        tree = ET.ElementTree(root)
        tree.write(path)



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math

from .sun_position import SunPosition


class SolarIrradiance:
    r"""
    A class to model the solar irradiance.
    """

    def __init__(self, sun_position: SunPosition):
        r"""
        To instantiate the ``SolarIrradiance`` class, provide the following parameter.


        :param sun_position: An instance of the SunPosition class.
        :type sun_position: pysolorie.SunPosition
        """
        self.sun_position = sun_position

    def calculate_extraterrestrial_irradiance(self, day_of_year: int) -> float:
        r"""
        Calculate the extraterrestrial solar irradiance for a given day of the year.

        The extraterrestrial solar irradiance, :math:`I`,
        is the annual average of the Sun's irradiation intensity.
        This refers to the amount of solar energy received per unit
        area per unit time on a surface perpendicular to the Sun's rays,
        outside Earth's atmosphere.

        The formula used is:

        .. math::
            I(n) = SC
            \times (1 + 0.033 \times \cos (\frac{2\pi~n}{365}))


        | - :math:`SC` is the average solar radiation arriving outside
           of the Earth's atmosphere,
           which is approximately ``1367`` Megawatts per square meter [1]_.
           This is also known as the solar constant.
        | - The factor ``0.033``, which is two times the eccentricity of
            the Earth's orbit around the Sun, accounts for the variation
            in the Earth-Sun distance due to the Earth's elliptical orbit.
        | - :math:`n` is the day of the year (i.e., ``day_of_year``)

        :param day\_of\_year: The day of the year, ranging from 1 to 365.
        :type day\_of\_year: int

        :return: The extraterrestrial solar irradiance in Megawatts per square meter.
        :rtype: float

        References
        ----------
        .. [1] Duffie (Deceased), J., Beckman, W., & Blair, N. (2020).
                Solar Engineering of Thermal Processes, Photovoltaics and Wind. Wiley.
        """

        # Solar constant (MW/m^2)
        SOLAR_CONSTANT = 1367 * 1e-6

        # Factor to account for the Earth's orbital eccentricity
        earth_orbital_eccentricity = 0.033

        # Calculate the extraterrestrial solar irradiance
        extraterrestrial_irradiance = SOLAR_CONSTANT * (
            1 + earth_orbital_eccentricity * math.cos(2 * math.pi * day_of_year / 365)
        )

        return extraterrestrial_irradiance



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Optional


class InvalidClimateTypeError(ValueError):
    r"""
    Exception raised for invalid climate type.

    :param climate_type: input climate type which caused the error
    :type climate_type: str
    :param message: explanation of the error, defaults to "Invalid climate type"
    :type message: str, optional
    """

    def __init__(self, climate_type: str, message: Optional[str] = None):
        self.VALID_CLIMATE_TYPES = [
            "TROPICAL",
            "MIDLATITUDE SUMMER",
            "SUBARCTIC SUMMER",
            "MIDLATITUDE WINTER",
        ]
        self.climate_type = climate_type
        if message is None:
            message = (
                f"Invalid climate type: {climate_type}. "
                f"Valid climate types are {', '.join(self.VALID_CLIMATE_TYPES)}"
            )
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        r"""
        String representation of the exception.

        :return: formatted string indicating the invalid
                 climate type and the error message
        :rtype: str
        """
        return f"{self.message}"


class MissingObserverLatitudeError(ValueError):
    r"""
    Exception raised when the observer's latitude is not provided.

    :param message: explanation of the error, defaults to "Missing required data:
                    Observer latitude. Please ensure to provide the latitude of
                    the observer."
    :type message: str, optional
    """

    def __init__(
        self,
        message: str = (
            "Missing required data: Observer latitude. "
            "Please ensure to provide the latitude of the observer."
        ),
    ):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        r"""
        String representation of the exception.

        :return: formatted string indicating the invalid
                 climate type and the error message
        :rtype: str
        """
        return f"{self.message}"


class InvalidObserverLatitudeError(ValueError):
    """
    Exception raised when the observer's latitude is not within the valid range.

    :param message: explanation of the error, defaults to
                    "Invalid data: Observer latitude.
                    Please ensure to provide a latitude between -88 and 88 degrees."
    :type message: str, optional
    """

    def __init__(
        self,
        message: str = (
            "Invalid data: Observer latitude. "
            "Please ensure to provide a latitude between -88 and 88 degrees."
        ),
    ):
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        """
        String representation of the exception.

        :return: formatted string indicating the invalid
                 latitude and the error message
        :rtype: str
        """
        return f"{self.message}"



# Copyright 2023 Alireza Aghamohammadi

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
from scipy import integrate, optimize  # type: ignore

from .atmospheric_transmission import AtmosphericTransmission
from .irradiance import SolarIrradiance
from .observer import Observer
from .sun_position import SunPosition


class IrradiationCalculator:
    r"""
    A class to find the optimal orientation and
    calculate the direct irradiation for a solar panel [1]_.

    References
    ----------
    .. [1] Aghamohammadi, A., & Foulaadvand, M. (2023).
            Efficiency comparison between tracking and
            optimally fixed flat solar collectors. Scientific Reports, 13(1).


    """

    OMEGA = 7.2722 * 1e-5

    def __init__(
        self, climate_type: str, observer_altitude: int, observer_latitude: float
    ):
        """
        To instantiate the ``IrradiationCalculator`` class,
        provide the following three parameters.

        :param climate_type: The type of climate.
        :type climate_type: str
        :param observer_altitude: The altitude of the observer in meters.
        :type observer_altitude: int
        :param observer_latitude: The latitude of the observer in degrees.
        :type observer_latitude: float
        """
        self._observer = Observer(observer_latitude=observer_latitude)
        self._sun_position = SunPosition()
        self._solar_irradiance = SolarIrradiance(self._sun_position)
        self._atmospheric_transmission = AtmosphericTransmission(
            climate_type, observer_altitude, observer_latitude
        )

    @staticmethod
    def _heaviside(x: float) -> int:
        """
        Heaviside step function.

        :param x: Input to the function.
        :type x: float
        :return: 1 if x >= 0, else 0.
        :rtype: int
        """
        return 1 if x >= 0 else 0

    def _calculate_irradiance_component(
        self, hour_angle: float, panel_orientation: float, day_of_year: int
    ) -> float:
        r"""
        Calculate a component of the irradiance integral.

        :param hour_angle: The hour angle.
        :type hour_angle: float
        :param panel_orientation: The orientation of the solar panel in radians.
        :type panel_orientation: float
        :param day_of_year: The day of the year.
        :type day_of_year: int
        :return: A component of the irradiance integral.
        :rtype: float
        """
        observer_latitude = self._observer._ensure_latitude_provided()
        solar_declination = self._sun_position.solar_declination(day_of_year)
        cos_theta = math.sin(solar_declination) * math.sin(
            observer_latitude - panel_orientation
        ) + math.cos(solar_declination) * math.cos(hour_angle) * math.cos(
            observer_latitude - panel_orientation
        )
        transmittance = self._atmospheric_transmission.calculate_transmittance(
            day_of_year, self._sun_position.solar_time(hour_angle)
        )
        irradiance = self._solar_irradiance.calculate_extraterrestrial_irradiance(
            day_of_year
        )
        return (
            irradiance
            * transmittance
            * cos_theta
            * self._heaviside(cos_theta)
            / self.OMEGA
        )

    def calculate_direct_irradiation(
        self, panel_orientation: float, day_of_year: int
    ) -> float:
        r"""
        Calculate the direct irradiation
        for a given solar panel orientation (i.e., :math:`\beta`).

        | The direct irradiation is calculated using the formula:

        .. math::
            E(n,\phi) = \frac{I(n)}{\Omega} \int_{\omega_s}^{\omega_t}
            \cos(\theta) \times H(\cos(\theta)) \times \tau_b~d\omega


        | - :math:`n` is the day of year (i.e., ``day_of_year``)

        | - :math:`\phi` is the latitude of the observer

        | - :math:`I(n)`  is the amount of
                solar energy received per unit area per second
                on day number :math:`n` of the year

        | - :math:`\Omega` = ``7.2722 * 1e-5``

        | - :math:`\theta` is incidence angle, the angle between the position vector
                of the sun and the normal vector to the solar panel.

        | - :math:`\omega_s` is the sunrise hour angle

        | - :math:`\omega_t` is the sunset hour angle

        | - :math:`H` is the Heaviside step function


        :param panel_orientation: The orientation of the solar panel in degrees.
        :type panel_orientation: float
        :param day_of_year: The day of the year.
        :type day_of_year: int
        :return: The direct irradiation in Megajoules per square meter.
        :rtype: float
        """
        sunrise_hour_angle, sunset_hour_angle = self._observer.calculate_sunrise_sunset(
            day_of_year
        )
        panel_orientation = math.radians(panel_orientation)
        irradiance_components = [
            self._calculate_irradiance_component(
                hour_angle, panel_orientation, day_of_year
            )
            for hour_angle in np.arange(sunrise_hour_angle, sunset_hour_angle, 0.01)
        ]

        # During polar night, the sun doesn't rise and both hour angles are zero.
        # This results in an empty irradiance_components list. In this case,
        # we return 0 as there is no direct irradiation.
        if not irradiance_components:
            return 0

        return integrate.simpson(irradiance_components, dx=0.01)

    def find_optimal_orientation(self, day_of_year: int) -> float:
        """
        Find the optimal orientation :math:`beta` that maximizes
        the direct irradiation.

        :param day_of_year: The day of the year.
        :type day_of_year: int
        :return: The optimal orientation (i.e., :math:`beta`) in degrees.
        :rtype: float
        """

        def neg_irradiation(beta: float):
            # We negate the irradiation because we're minimizing
            return -self.calculate_direct_irradiation(math.degrees(beta), day_of_year)

        result = optimize.minimize_scalar(
            neg_irradiation, bounds=(-math.pi / 2, math.pi / 2), method="bounded"
        )
        optimal_beta = result.x

        # Check if irradiance_components is empty
        if not self.calculate_direct_irradiation(
            math.degrees(optimal_beta), day_of_year
        ):
            observer_latitude = self._observer._ensure_latitude_provided()
            # Return 90 or -90 based on observer_latitude
            return 90 if observer_latitude >= 0 else -90

        return math.degrees(optimal_beta)



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging
from pathlib import Path
from typing import Any, Dict, List
from unittest.mock import MagicMock

import matplotlib.pyplot as plt
import numpy as np

from pysolorie import IrradiationCalculator, Plotter


def test_plot_optimal_orientation(caplog, tmpdir) -> None:
    caplog.set_level(logging.INFO)

    # Create a temporary directory for the test
    temp_dir: Path = Path(tmpdir)

    # Initialize the Plotter
    plotter: Plotter = Plotter()

    # Initialize the IrradiationCalculator for Tehran
    irradiation_calculator: IrradiationCalculator = IrradiationCalculator(
        "MIDLATITUDE SUMMER", 1200, 35.6892
    )

    # Define the path for the plot
    plot_path: Path = temp_dir / "plot.png"
    from_day: int = 60
    to_day: int = 70
    # Call the method to generate the plot
    plotter.plot_optimal_orientation(
        irradiation_calculator, from_day, to_day, plot_path
    )

    # Check the plot file
    assert plot_path.exists(), "The plot file was not created."

    img: np.ndarray = plt.imread(plot_path)
    assert img.shape[0] > 0, "The plot image has no content."
    assert img.shape[1] > 0, "The plot image has no content."

    # Check the logs
    for day in range(from_day, to_day):
        assert any(
            f"On day {day}," in record.message for record in caplog.records
        ), f"No log message for day {day}"


def test_plot_total_direct_irradiation(caplog, tmpdir) -> None:
    caplog.set_level(logging.INFO)

    # Create a temporary directory for the test
    temp_dir: Path = Path(tmpdir)

    # Initialize the Plotter
    plotter: Plotter = Plotter()

    # Initialize the IrradiationCalculator for Tehran
    irradiation_calculator: IrradiationCalculator = IrradiationCalculator(
        "MIDLATITUDE SUMMER", 1200, 35.6892
    )

    # Define the path for the plot
    plot_path: Path = temp_dir / "plot.png"
    from_day: int = 60
    to_day: int = 70
    # Call the method to generate the plot
    plotter.plot_total_direct_irradiation(
        irradiation_calculator, from_day, to_day, plot_path
    )

    # Check the plot file
    assert plot_path.exists(), "The plot file was not created."

    img: np.ndarray = plt.imread(plot_path)
    assert img.shape[0] > 0, "The plot image has no content."
    assert img.shape[1] > 0, "The plot image has no content."

    # Check the logs
    for day in range(from_day, to_day):
        assert any(
            f"On day {day}," in record.message for record in caplog.records
        ), f"No log message for day {day}"


def test_plot_method() -> None:
    # Set up the necessary variables
    plotter: Plotter = Plotter()
    days: List[int] = [1, 2, 3]
    betas: List[float] = [10.0, 20.0, 30.0]
    path: None = None
    plot_kwargs: Dict[str, Any] = {}
    savefig_kwargs: Dict[str, Any] = {}

    # Replace plt.show with a mock
    plt.show = MagicMock()

    # Call the method
    plotter._plot(days, betas, path, plot_kwargs, savefig_kwargs)
    plt.show.assert_called_once()



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from typing import Tuple

import pytest

from pysolorie import HottelModel, exceptions


@pytest.mark.parametrize(
    "climate_type, observer_altitude, expected_result",
    [
        ("MIDLATITUDE SUMMER", 1200, (0.228, 0.666, 0.309)),  # Tehran Summer
        ("MIDLATITUDE WINTER", 1200, (0.242, 0.679, 0.303)),  # Tehran Winter
        ("TROPICAL", 26, (0.124, 0.739, 0.392)),  # Medan
        ("SUBARCTIC SUMMER", 136, (0.140, 0.739, 0.379)),  # Fairbanks
    ],
)
def test_calculate_transmittance_components(
    climate_type: str,
    observer_altitude: int,
    expected_result: Tuple[float, float, float],
) -> None:
    hottel_model: HottelModel = HottelModel()
    result: Tuple[float, float, float] = (
        hottel_model.calculate_transmittance_components(climate_type, observer_altitude)
    )
    assert pytest.approx(result, abs=1e-3) == expected_result


def test_invalid_climate_type() -> None:
    with pytest.raises(ValueError, match="Invalid climate type"):
        hottel_model: HottelModel = HottelModel()
        hottel_model.calculate_transmittance_components("INVALID", 1000)


def test_invalid_climate_type_with_custom_message() -> None:
    custom_message = "Custom error message"
    with pytest.raises(exceptions.InvalidClimateTypeError, match=custom_message):
        raise exceptions.InvalidClimateTypeError("INVALID", custom_message)



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math

import pytest

from pysolorie import SunPosition


@pytest.mark.parametrize(
    "day_of_year, solar_time, expected_declination, expected_hour_angle",
    [
        (1, 12 * 60 * 60, -0.4014257279586958, 0),  # January 1st at noon
        (81, 10 * 60 * 60, 0, -math.pi / 6),  # March 22nd at 10am (equinox)
        (81, 12 * 60 * 60, 0, 0),  # March 22nd at noon (equinox)
        (1, 13 * 60 * 60, -0.4014257279586958, math.pi / 12),  # January 1st at 1pm
    ],
)
def test_sun_position(
    day_of_year: int,
    solar_time: int,
    expected_declination: float,
    expected_hour_angle: float,
) -> None:
    sun_position: SunPosition = SunPosition()
    declination: float = sun_position.solar_declination(day_of_year)
    hour_angle: float = sun_position.hour_angle(solar_time)
    assert declination == pytest.approx(expected_declination, abs=1e-3)
    assert hour_angle == pytest.approx(expected_hour_angle, abs=1e-3)


@pytest.mark.parametrize(
    "hour_angle, expected_solar_time",
    [
        (0, 12 * 60 * 60),  # solar noon
        (-math.pi / 6, 10 * 60 * 60),  # 10am
        (math.pi / 12, 13 * 60 * 60),  # 1pm
        (math.pi, 24 * 60 * 60),  # solar night
    ],
)
def test_solar_time(
    hour_angle: float,
    expected_solar_time: int,
) -> None:
    sun_position: SunPosition = SunPosition()
    solar_time: float = sun_position.solar_time(hour_angle)
    assert solar_time == pytest.approx(expected_solar_time, abs=1e-3)



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import functools
import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


def logger_decorator(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self.logger = logging.getLogger(func.__name__)
        self.logger.setLevel(logging.INFO)
        result = func(self, *args, **kwargs)
        return result

    return wrapper



# Copyright 2023 Alireza Aghamohammadi

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Dict, List, Optional, Tuple

import matplotlib.pyplot as plt  # type: ignore

from .logger import logger_decorator
from .numerical_integration import IrradiationCalculator


class Plotter:
    r"""
    A class used to plot the optimal orientation of a solar panel.
    """

    @logger_decorator
    def plot_optimal_orientation(
        self,
        irradiation_calculator: IrradiationCalculator,
        from_day: int,
        to_day: int,
        path: Optional[Path] = None,
        plot_kwargs: Optional[Dict[str, str]] = None,
        savefig_kwargs: Optional[Dict[str, str]] = None,
    ) -> None:
        r"""
        Plots the optimal orientation of a solar panel for a range of days.

        :param irradiation_calculator: An instance of the IrradiationCalculator class.
        :type irradiation_calculator: pysolorie.IrradiationCalculator
        :param from_day: The starting day for the range of days.
        :type from_day: int
        :param to_day: The ending day for the range of days.
        :type to_day: int
        :param path: The path where the plot will be saved (default is None,
                     which means the plot will be shown but not saved).
        :type path: Path, optional
        :param plot_kwargs: A dictionary of keyword arguments
                            to be passed to the plot (default is None).
        :type plot_kwargs: dict, optional
        :param savefig_kwargs: A dictionary of keyword arguments
                               to be passed to the savefig function (default is None).
        :type savefig_kwargs: dict, optional
        """

        days, betas = self._calculate_optimal_orientations(
            irradiation_calculator, from_day, to_day
        )

        plot_kwargs = plot_kwargs if plot_kwargs else {}
        savefig_kwargs = savefig_kwargs if savefig_kwargs else {}

        self._plot(days, betas, path, plot_kwargs, savefig_kwargs)

    @logger_decorator
    def plot_total_direct_irradiation(
        self,
        irradiation_calculator: IrradiationCalculator,
        from_day: int,
        to_day: int,
        path: Optional[Path] = None,
        plot_kwargs: Optional[Dict[str, str]] = None,
        savefig_kwargs: Optional[Dict[str, str]] = None,
    ) -> None:
        r"""
        Plots the total direct irradiation for a range of days.

        :param irradiation_calculator: An instance of the IrradiationCalculator class.
        :type irradiation_calculator: pysolorie.IrradiationCalculator
        :param from_day: The starting day for the range of days.
        :type from_day: int
        :param to_day: The ending day for the range of days.
        :type to_day: int
        :param path: The path where the plot will be saved (default is None,
                    which means the plot will be shown but not saved).
        :type path: Path, optional
        :param plot_kwargs: A dictionary of keyword arguments
                            to be passed to the plot (default is None).
        :type plot_kwargs: dict, optional
        :param savefig_kwargs: A dictionary of keyword arguments
                            to be passed to the savefig function (default is None).
        :type savefig_kwargs: dict, optional
        """

        days, betas = self._calculate_optimal_orientations(
            irradiation_calculator, from_day, to_day
        )
        total_direct_irradiations = [
            irradiation_calculator.calculate_direct_irradiation(beta, day)
            for day, beta in zip(days, betas)
        ]

        plot_kwargs = plot_kwargs if plot_kwargs else {}
        savefig_kwargs = savefig_kwargs if savefig_kwargs else {}

        self._plot(days, total_direct_irradiations, path, plot_kwargs, savefig_kwargs)

    @logger_decorator
    def _calculate_optimal_orientations(
        self, irradiation_calculator: IrradiationCalculator, from_day: int, to_day: int
    ) -> Tuple[List[int], List[float]]:
        days = []
        betas = []

        for day in range(from_day, to_day):
            beta = irradiation_calculator.find_optimal_orientation(day)
            self.logger.info(  # type: ignore
                f"On day {day},"
                + f"the solar panel's optimal orientation is {beta} degrees."
            )
            days.append(day)
            betas.append(beta)

        return days, betas

    def _plot(
        self,
        days: List[int],
        betas: List[float],
        path: Optional[Path],
        plot_kwargs: Dict[str, str],
        savefig_kwargs: Dict[str, str],
    ) -> None:
        figsize = plot_kwargs.get("figsize", (10, 6))
        _, ax = plt.subplots(figsize=figsize)
        ax.plot(days, betas)
        ax.set_xlabel(plot_kwargs.get("xlabel", "X Axis Title"))
        ax.set_ylabel(plot_kwargs.get("ylabel", "Y Axis Title"))
        ax.set_title(plot_kwargs.get("title", "Title"))
        ax.grid(True)

        if path is not None:
            plt.savefig(path, **savefig_kwargs)
        else:
            plt.show()



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pytest

from pysolorie import IrradiationCalculator


@pytest.mark.parametrize(
    "climate_type,"
    + "observer_altitude,"
    + "observer_latitude,"
    + "day_of_year,"
    + "panel_orientation,"
    + "expected_result",
    [
        (
            "MIDLATITUDE SUMMER",
            1200,
            35.6892,
            172,
            45.0,
            19.961,  # Tehran Summer, day_of_year=172 (June 21)
        ),
        (
            "MIDLATITUDE WINTER",
            1200,
            35.6892,
            355,
            45.0,
            19.110,  # Tehran Winter, day_of_year=355 (Dec 21)
        ),
        (
            "TROPICAL",
            26,
            3.5952,
            100,
            45.0,
            13.003,  # Medan, day_of_year=100 (April 10)
        ),
        (
            "SUBARCTIC SUMMER",
            132,
            64.84361,
            200,
            45.0,
            21.012,  # Fairbanks Summer, day_of_year=200 (July 19)
        ),
        (
            "SUBARCTIC SUMMER",
            356,
            70.00,
            355,
            45.0,
            0,  # Finnmark Winter, day_of_year=355 (December 21)
        ),
    ],
)
def test_calculate_direct_irradiation(
    climate_type: str,
    observer_altitude: int,
    observer_latitude: float,
    day_of_year: int,
    panel_orientation: float,
    expected_result: float,
) -> None:
    irradiation_calculator = IrradiationCalculator(
        climate_type, observer_altitude, observer_latitude
    )
    result = irradiation_calculator.calculate_direct_irradiation(
        panel_orientation, day_of_year
    )
    assert pytest.approx(result, abs=1e-3) == expected_result


@pytest.mark.parametrize(
    "climate_type, observer_altitude, observer_latitude, day_of_year, expected_result",
    [
        (
            "MIDLATITUDE SUMMER",
            1200,
            35.6892,
            172,
            0.170,
        ),  # Tehran Summer, day_of_year=172 (June 21)
        (
            "MIDLATITUDE WINTER",
            1200,
            35.6892,
            355,
            63.791,
        ),  # Tehran Winter, day_of_year=355 (Dec 21)
        (
            "TROPICAL",
            26,
            3.5952,
            100,
            -6.610,
        ),  # Medan, day_of_year=100 (April 10)
        (
            "SUBARCTIC SUMMER",
            132,
            64.84361,
            200,
            32.614,
        ),  # Fairbanks Summer, day_of_year=200 (July 19)
        (
            "SUBARCTIC SUMMER",
            356,
            70.00,
            355,
            90,  # Finnmark Winter, day_of_year=355 (December 21)
        ),
        (
            "SUBARCTIC SUMMER",
            356,
            70.00,
            172,
            31.505,  # Finnmark Summer, day_of_year=172 (June 21)
        ),
    ],
)
def test_find_optimal_orientation(
    climate_type: str,
    observer_altitude: int,
    observer_latitude: float,
    day_of_year: int,
    expected_result: float,
) -> None:
    irradiation_calculator = IrradiationCalculator(
        climate_type, observer_altitude, observer_latitude
    )
    result = irradiation_calculator.find_optimal_orientation(day_of_year)
    assert pytest.approx(result, abs=1e-3) == expected_result



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import pytest

from pysolorie import SolarIrradiance, SunPosition


@pytest.mark.parametrize(
    "day_of_year, expected_irradiance",
    [
        (1, 0.001411444),  # January 1st
        (81, 0.001374918),  # March 22nd (equinox)
        (172, 0.00132262),  # June 21st (summer solstice)
        (264, 0.001359464),  # September 23rd (equinox)
        (355, 0.001412104),  # December 21st (winter solstice)
    ],
)
def test_calculate_extraterrestrial_irradiance(
    day_of_year: int, expected_irradiance: float
) -> None:
    sun_position = SunPosition()
    solar_irradiance = SolarIrradiance(sun_position)
    irradiance: float = solar_irradiance.calculate_extraterrestrial_irradiance(
        day_of_year
    )
    assert irradiance == pytest.approx(expected_irradiance, abs=1e-5)



#    Copyright 2023 Alireza Aghamohammadi

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math

import pytest

from pysolorie import Observer


@pytest.mark.parametrize(
    "observer_latitude,"
    + "observer_longitude,"
    + "day_of_year,"
    + "solar_time,"
    + "expected_zenith_angle",
    [
        (
            35.69,
            51.39,
            81,
            12 * 60 * 60,
            0.623,
        ),  # Test at Tehran on the 81st day of the year at noon
        (
            35.69,
            51.39,
            355,
            12 * 60 * 60,
            1.032,
        ),  # Test at Tehran on the 355th day of the year at noon
        (
            3.59,
            98.67,
            81,
            10 * 60 * 60,
            0.527,
        ),  # Test at Medan on the 81st day of the year at 10am
        (
            64.84,
            -147.72,
            1,
            13 * 60 * 60,
            1.547,
        ),  # Test at Fairbanks on the 1st day of the year at 1pm
    ],
)
def test_calculate_zenith_angle(
    observer_latitude: float,
    observer_longitude: float,
    day_of_year: int,
    solar_time: int,
    expected_zenith_angle: float,
) -> None:
    observer: Observer = Observer(observer_latitude, observer_longitude)
    expected_observer_latitude = math.radians(observer_latitude)
    expected_observer_longitude = math.radians(observer_longitude)

    assert observer.observer_latitude == pytest.approx(
        expected_observer_latitude, abs=1e-3
    )
    assert observer.observer_longitude == pytest.approx(
        expected_observer_longitude, abs=1e-3
    )

    zenith_angle: float = observer.calculate_zenith_angle(day_of_year, solar_time)
    assert zenith_angle == pytest.approx(expected_zenith_angle, abs=1e-3)


def test_calculate_zenith_angle_without_latitude():
    observer = Observer(None, 0)
    with pytest.raises(
        ValueError,
        match="Missing required data: Observer latitude",
    ):
        observer.calculate_zenith_angle(1, 12 * 60 * 60)


@pytest.mark.parametrize(
    "observer_latitude,"
    + "day_of_year,"
    + "expected_sunrise_hour_angle,"
    + "expected_sunset_hour_angle",
    [
        (35.69, 1, -1.261, 1.261),  # January 1st, Tehran
        (35.69, 81, -math.pi / 2, math.pi / 2),  # March 22nd (equinox), Tehran
        (35.69, 172, -1.888, 1.888),  # June 21st (solstice), Tehran
        (70.00, 1, 0, 0),  # January 1st, Finnmark
        (70.00, 81, -math.pi / 2, math.pi / 2),  # March 22nd (equinox), Finnmark
        (70.00, 172, -math.pi, math.pi),  # June 21st (solstice), Finnmark
    ],
)
def test_calculate_sunrise_sunset(
    observer_latitude: float,
    day_of_year: int,
    expected_sunrise_hour_angle: float,
    expected_sunset_hour_angle: float,
) -> None:
    observer: Observer = Observer(observer_latitude=observer_latitude)
    sunrise_hour_angle, sunset_hour_angle = observer.calculate_sunrise_sunset(
        day_of_year
    )
    assert sunrise_hour_angle == pytest.approx(expected_sunrise_hour_angle, abs=1e-3)
    assert sunset_hour_angle == pytest.approx(expected_sunset_hour_angle, abs=1e-3)


def test_calculate_sunrise_sunset_invalid_latitude():
    invalid_latitudes = [89, -89]
    for invalid_latitude in invalid_latitudes:
        observer = Observer(invalid_latitude)
        with pytest.raises(
            ValueError,
            match="Invalid data: Observer latitude. ",
        ):
            observer.calculate_sunrise_sunset(day_of_year=172)



# Copyright 2023 Alireza Aghamohammadi

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Optional

from .exceptions import InvalidObserverLatitudeError, MissingObserverLatitudeError
from .sun_position import SunPosition


class Observer:
    r"""
    A class to model an observer based on horizontal and equatorial pictures
    of the sun-earth geometry.
    """

    def __init__(
        self,
        observer_latitude: Optional[float] = None,
        observer_longitude: Optional[float] = None,
    ):
        """
        To instantiate the ``Observer`` class, provide the following parameter.

        :param observer_latitude: The latitude of the observer in degrees (optional).
        :type observer_latitude: Optional[float]
        :param observer_longitude: The longitude of the observer in degrees (optional).
        :type observer_longitude: Optional[float]
        """
        self.observer_latitude = observer_latitude
        self.observer_longitude = observer_longitude
        self.sun_position = SunPosition()

    def calculate_zenith_angle(self, day_of_year: int, solar_time: float) -> float:
        r"""
        Calculate the solar zenith angle.

        The solar zenith angle is calculated using the formula:

        .. math::
            \cos(\theta_z) = \sin(\phi) \times \sin(\delta)
            + \cos(\phi) \times \cos(\delta) \times \cos(\omega)


        | - :math:`\theta_z` is the solar zenith angle
        | - :math:`\phi` is the latitude of the observer
        | - :math:`\delta` is the solar declination
        | - :math:`\omega` is the hour angle.

        :param day_of_year: The day of the year.
        :type day_of_year: int
        :param solar_time: The solar time in seconds.
        :type solar_time: float
        :return: The zenith angle in radians.
        :rtype: float
        """
        observer_latitude = self._ensure_latitude_provided()

        solar_declination = self.sun_position.solar_declination(day_of_year)
        hour_angle = self.sun_position.hour_angle(solar_time)
        return math.acos(
            math.sin(observer_latitude) * math.sin(solar_declination)
            + math.cos(observer_latitude)
            * math.cos(solar_declination)
            * math.cos(hour_angle)
        )

    def calculate_sunrise_sunset(self, day_of_year: int) -> tuple:
        r"""
        Calculate the hour angle at sunrise and sunset.

        The hour angle at sunrise and sunset is calculated using the formula:

        .. math::
            \cos(\omega) = -\tan(\phi) \times \tan(\delta)


        | - :math:`\omega` is the hour angle
        | - :math:`\phi` is the latitude of the observer
        | - :math:`\delta` is the solar declination.

        :param day_of_year: The day of the year.
        :type day_of_year: int
        :return: The hour angle at sunrise and sunset in radians.
        :rtype: tuple
        """
        observer_latitude = self._ensure_latitude_provided()

        solar_declination = self.sun_position.solar_declination(day_of_year)

        tan_product = -math.tan(observer_latitude) * math.tan(solar_declination)

        if tan_product > 1:
            return 0, 0

        if tan_product < -1:
            return -math.pi, math.pi

        hour_angle = math.acos(tan_product)

        sunrise = -hour_angle
        sunset = hour_angle

        return sunrise, sunset

    def _ensure_latitude_provided(self) -> float:
        if self.observer_latitude is None:
            raise MissingObserverLatitudeError()
        if not math.radians(-88) <= self.observer_latitude <= math.radians(88):
            raise InvalidObserverLatitudeError()
        return self.observer_latitude

    _observer_latitude: Optional[float]

    @property
    def observer_latitude(self) -> Optional[float]:
        """
        Getter for the observer's latitude.

        :return: The observer's latitude in radians.
        :rtype: Optional[float]
        """
        return self._observer_latitude

    @observer_latitude.setter
    def observer_latitude(self, value: Optional[float]):
        """
        Setter for the observer's latitude. Converts the value to radians if not None.

        :param value: The observer's latitude in degrees.
        :type value: Optional[float]
        """
        self._observer_latitude = math.radians(value) if value is not None else None

    _observer_longitude: Optional[float]

    @property
    def observer_longitude(self) -> Optional[float]:
        """
        Getter for the observer's longitude.

        :return: The observer's longitude in radians.
        :rtype: Optional[float]
        """
        return self._observer_longitude

    @observer_longitude.setter
    def observer_longitude(self, value: Optional[float]):
        """
        Setter for the observer's longitude. Converts the value to radians if not None.

        :param value: The observer's longitude in degrees.
        :type value: Optional[float]
        """
        self._observer_longitude = math.radians(value) if value is not None else None
